from idaes.core.base.property_base import PhysicalParameterBlock
from property_packages.modular.modular_extended import GenericExtendedParameterBlock
from ahuora_builder_types.flowsheet_schema import PropertyPackageType, CustomCompoundPropertiesSchema
from idaes.core import LiquidPhase, VaporPhase, Component
from pyomo.environ import units as pyunits
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import (
    CubicComplementarityVLE,
)
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity
from idaes.models.properties.modular_properties.pure import RPP4






def build_custom_package(property_package_schema: PropertyPackageType) -> PhysicalParameterBlock:
    """
    Build a custom property package from a schema
    """
    if property_package_schema.custom_compound_properties is None:
        raise ValueError("Custom compound properties must be provided for custom property package")
    if property_package_schema.custom_package_properties is None:
        raise ValueError("Custom package properties must be provided for custom property package")
    if property_package_schema.custom_kappa_values is None:
        raise ValueError("Custom kappa values must be provided for custom property package")
    

    pp = property_package_schema.custom_package_properties
    # assert nothing in pp is None
    for key, value in pp.model_dump().items():
        if value is None:
            raise ValueError(f"Custom package property {key} must be provided for custom property package")

    return GenericExtendedParameterBlock(
        **{
            "components": build_components(property_package_schema.compounds,property_package_schema.custom_compound_properties),
                # Specifying phases
            "phases": {
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
                "Vap": {
                    "type": VaporPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
            },
            # Set base units of measurement
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            # Specifying state definition
            "state_definition": FTPx,
            "state_bounds": {
                "flow_mol": (pp.flow_mol_min, pp.flow_mol_nominal, pp.flow_mol_max, pyunits.mol / pyunits.s),
                "temperature": (pp.temperature_min, pp.temperature_nominal, pp.temperature_max, pyunits.K),
                "pressure": (pp.pressure_min, pp.pressure_nominal, pp.pressure_max, pyunits.Pa),
            },
            "pressure_ref": (pp.pressure_ref, pyunits.Pa),
            "temperature_ref": (pp.temperature_ref, pyunits.K),
            # Defining phase equilibria
            "phases_in_equilibrium": [("Vap", "Liq")],
            "phase_equilibrium_state": {("Vap", "Liq"): CubicComplementarityVLE},
            "bubble_dew_method": LogBubbleDew,
            "parameter_data": {
                "PR_kappa": build_kappas(property_package_schema.custom_kappa_values)
            },
        }
    )



def build_components(compounds: list[str], custom_compound_properties: list[CustomCompoundPropertiesSchema]) -> dict[str, dict]:
    compound_configuration: dict[str,dict] = {}
    for compound, properties in zip(compounds, custom_compound_properties):
        compound_configuration[compound] = {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (properties.mw, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (properties.pressure_crit, pyunits.Pa),  # [1]
                "temperature_crit": (properties.temperature_crit, pyunits.K),  # [1]
                "omega": properties.omega,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (properties.cp_mol_ig_comp_coeff_A, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (properties.cp_mol_ig_comp_coeff_B, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (properties.cp_mol_ig_comp_coeff_C, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (properties.cp_mol_ig_comp_coeff_D, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (properties.enth_mol_form_vap_comp_ref, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (properties.entr_mol_form_vap_comp_ref, pyunits.J / pyunits.mol / pyunits.K),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (properties.pressure_sat_comp_coeff_A, None),  # [1]
                    "B": (properties.pressure_sat_comp_coeff_B, None),
                    "C": (properties.pressure_sat_comp_coeff_C, None),
                    "D": (properties.pressure_sat_comp_coeff_D, None),
                },
            },
        }
    return compound_configuration


def build_kappas(kappa_values: dict[str,dict[str,float]]) -> dict[tuple[str,str], float]:
    kappa_configuration: dict[tuple[str,str], float] = {}
    for comp1, values in kappa_values.items():
        for comp2, value in values.items():
            kappa_configuration[(comp1, comp2)] = value
    return kappa_configuration


from pyomo.environ import ConcreteModel, Var, value, Reference, Expression, Constraint
from ..property_map_manipulation import update_property, free_property
from ..adapter import fix_slice
from ...properties_manager import PropertiesManager
from ahuora_builder_types.id_types import PropertyValueId

from idaes.core.util.model_statistics import degrees_of_freedom

def test_free_property():
    m = ConcreteModel()
    m.x = Var(initialize=2)
    # The extra variable and constraint here is just to make it a proper model.
    m.y = Var(initialize=3)
    m.constraint = Constraint(expr=m.x + m.y == 10)
    m.expr = Expression(expr=m.x + 1)
    properties_map = PropertiesManager()
    properties_map.add(
        PropertyValueId(123), Reference(m.x), "m.x"
    )
    properties_map.add(
        PropertyValueId(101), Reference(m.expr), "m.expr"
    )

    assert degrees_of_freedom(m) == 1
    property_component = properties_map.get(PropertyValueId(123))
    constraints = fix_slice(property_component.component, [5])
    properties_map.add_constraint(PropertyValueId(123), constraints)
    assert degrees_of_freedom(m) == 0

    # Test with fixing and unfixing a variable
    assert m.x.is_fixed()
    free_property(property_component)
    assert degrees_of_freedom(m) == 1
    assert not m.x.is_fixed()

    # Test with fixing and unfixing an expression
    expr_component = properties_map.get(PropertyValueId(101))
    constraints = fix_slice(expr_component.component, [10])
    properties_map.add_constraint(PropertyValueId(101), constraints)
    assert degrees_of_freedom(m) == 0
    free_property(expr_component)
    assert degrees_of_freedom(m) == 1




    


from __future__ import annotations
from pyomo.core.base.var import ScalarVar
from pyomo.environ import units as pyunits
from ahuora_builder.methods.adapter import fix_slice, fix_var
from ahuora_builder.properties_manager import PropertyComponent

from pyomo.core.base.expression import ScalarExpression, Expression
from pyomo.core.base.var import ScalarVar, VarData,  _GeneralVarData
from pyomo.core.base.constraint import ScalarConstraint, ConstraintData, _GeneralConstraintData
from pyomo.core.base.indexed_component import IndexedComponent
from pyomo.environ import value
from sympy import var
from ahuora_builder.methods.adapter import fix_slice
from .units_handler import ValueWithUnits




def update_property(property_component: PropertyComponent, values: list[ValueWithUnits]):
    """
    This function is used to update what a property component is fixed to.
    For example, in the online solving, we want to update properties to their new values from the OPC UA server, and then re-solve the model.
    This saves us from having to re-create the model and re-fix all the properties every time.
    Args:
    - property_component (PropertyComponent): The property component to update. This can be retrieved from the PropertiesManager.
    - values (list[ValueWithUnits]): The new values to set the property component to. These could be a list of values with the IDAES unit attached too.

    """
    constraints: list[ScalarVar | ScalarConstraint] = list(property_component.corresponding_constraint)
    if len(values) != len(constraints):
        raise ValueError(f"Number of values provided ({len(values)}) does not match the number of constraints ({len(constraints)}) for property {property_component.name} (id {property_component}). Please provide the correct number of values.")
    
    free_property(property_component)

    new_constraints = fix_slice(property_component.component, values)
    property_component.corresponding_constraint = new_constraints



def free_property(property_component: PropertyComponent):
    components: list[ScalarVar | ScalarExpression] = list(property_component.component.values())
    constraints: list[ScalarVar | ScalarConstraint] = list(property_component.corresponding_constraint)
   
    if len(components) != len(constraints):
        raise ValueError(f"Number of components ({len(components)}) does not match the number of corresponding constraints ({len(constraints)}) for property {property_component.name} (id {property_component}). This is likely an issue with the model setup internal to ahuora-builder, please check how the property and its corresponding constraint are added to the model.")



    for component, constraint in zip(components, constraints):
        if not (isinstance(component, ScalarVar) or isinstance(component, ScalarExpression)):
            # If this gets raised we have a problem and need to look at the get_variable_from_property function
            raise ValueError(f"Something is wrong, we should have either a scalar variable or expression for this property_component, got {type(component)} for {component.name}")

        if isinstance(component, ScalarVar):
            if not component.is_fixed():
                raise ValueError(f"Variable {component.name} must be fixed before applying input value. It looks like this variable is not fixed in the flowsheet scenario. Please choose variables that are fixed, or update the scenario JSON")
            component.unfix()
        else: # ScalarExpression
            if constraint is None:
                raise ValueError(f"Could not find corresponding constraint for expression property {property_component.name}. This means that the expression is not an input. please choose properties that are fixed/inputs, or update the scenario JSON so that {component.name} is an input.")
            # Remove the constraint(s), so that we can add a new one for the OPC input value
            try:
                blk = constraint.parent_block()
                print(f"[DEL] Removing constraint: {constraint.name}")
                blk.del_component(constraint)
            except Exception as e:
                print(f"[ERR] Could not delete constraint {constraint} (for {property_component.name}): {e}")
                raise e
            
    